local Lib = MissionMinder;
Lib.callbacks= Lib.callbacks or LibStub("CallbackHandler-1.0"):New(Lib)

local options = Lib.options;
local DEBUG = Lib.debug;

Lib.planner = {};
local module = Lib.planner;

local L = LibStub("AceLocale-3.0"):GetLocale("MissionMinder");
local L_1000 = L["K"]
local L_COMPLETED = L["completed"]
local L_WORKING = L["working"]
local L_IN_PARTY = L["in party"]
local L_AVAILABLE = L["available"]
local L_PARTY = L["Party"]
local L_COUNTERS = L["Counters"]
local L_NO_COUNTERS_AVAILABLE = L["No counters available"]
local L_ASSIGNED_MISSION = L["Assigned mission"]
MISSIONMINDER_MINIMIZE_MISSION = L["Minimize mission"];


local RED_BIAS_PERCENTAGE = 0.10
local ORANGE_BIAS_PERCENTAGE = 0.50;
local WHITE_BIAS_PERCENTAGE = 1.00;

local LEVELUP_ARROW_QUALITIES = {
	[LE_ITEM_QUALITY_UNCOMMON]	= "loottoast-arrow-green",
	[LE_ITEM_QUALITY_RARE] = "loottoast-arrow-blue",
	[LE_ITEM_QUALITY_EPIC] = "loottoast-arrow-purple",
	[LE_ITEM_QUALITY_LEGENDARY] = "loottoast-arrow-orange"
}



local minimizedParties = {} -- [missionID][1..3] = followerID
local followerMissions = {} -- [followerID] = mission

local function createInlineTexture(texture)
	return texture and ("|T" .. texture .. ":0|t") or nil;
end

local function canBeSentOnMission(follower)
	return follower and follower.isCollected and follower.status ~= GARRISON_FOLLOWER_INACTIVE;
end

local function isOnMission(follower)
	return follower and follower.status == GARRISON_FOLLOWER_ON_MISSION;
end

local function contains(list, value)
	if (list) then
		for _,v in pairs(list) do
			if(v == value) then
				return true
			end
		end
	end
	return false
end

local function isEmpty(list)
	if (list) then
		for _,v in pairs(list) do
			if ( v ) then
				return false
			end
		end
	end
	return true
end


local TRAIT_EXTRA_TRAINING = 80 -- whole party
local TRAIT_FAST_LEANER = 29 -- single follower
local TRAIT_HEARTHSTONE_PRO = 236 -- whole party

local expGainTraits = {
	[TRAIT_EXTRA_TRAINING] = 0.30,
	[TRAIT_FAST_LEANER] = 0.50,
	[TRAIT_HEARTHSTONE_PRO] = 0.35
}

local function ignoreTrait(traitID, hasBonusXPReward)
	if ( options.IgnoreXPGainTraits() and not hasBonusXPReward and expGainTraits[traitID]) then
		return true
	else
		return false
	end
end

local function followerWillMaxOnLevelUp(level, quality)
	local maxLevel = GarrisonMissionFrame.followerMaxLevel;
	local maxQuality = GarrisonMissionFrame.followerMaxQuality;

	local alreadyMaxLevel = level >= maxLevel
	local alreadyMaxQuality = quality >= maxQuality

	return (alreadyMaxQuality and level + 1 >= maxLevel) or (alreadyMaxLevel and quality + 1 >= maxQuality);
end



local function getMissionRewards(mission)
	local rewards = C_Garrison.GetMissionRewardInfo(mission.missionID);

	local t = {};
	for _, reward in pairs(rewards) do
		if(reward.followerXP) then
			local xp = reward.followerXP;
			if ( xp >= 1000 ) then
				xp = xp/1000 .. L_1000;
			end
			tinsert(t, "+" .. xp .. " " .. createInlineTexture(reward.icon));
		elseif (reward.itemID) then
			local itemTexture = select(10, GetItemInfo(reward.itemID));
			if(reward.quantity == 1) then
				tinsert(t, createInlineTexture(itemTexture));
			else
				tinsert(t, reward.quantity .. "x " .. createInlineTexture(itemTexture));
			end
		elseif(reward.currencyID) then
			if (reward.currencyID == 0) then -- gold
				tinsert(t, GetCoinTextureString(reward.quantity));
			else
				tinsert(t, reward.quantity .. " " .. createInlineTexture(reward.icon));
			end
		else
			-- because this is tooltip based, we ignore the value instead of writing the name
		end
	end

	return table.concat(t, ", ");
end

local function MissionHasBonusXPReward(missionID)
	local rewards = C_Garrison.GetMissionRewardInfo(missionID);
	for _, reward in pairs(rewards) do
		if(reward.followerXP) then
			return true;
		end
	end

	return false;
end

local timeLeft = {};
local inProgressMissions = {};
local function updateFollowerMissionTimeLeft(timeLeft)
	if ( MissionMinder.ptr ) then
		return;
	end

	C_Garrison.GetInProgressMissions(inProgressMissions);

	if (inProgressMissions and #inProgressMissions > 0) then
		for _,mission in pairs(inProgressMissions) do
			for _,id in pairs(mission.followers) do
				-- Old values are overwritten, but if a follower is AVAILABLE any previous values will still be present.
				--Thats Okay because we only use this data if the follower is GARRISON_FOLLOWER_ON_MISSION.
				timeLeft[id] = {
					["missionID"] = mission.missionID,
					["remaining"] = mission.timeLeft;
				};
			end
		end
	end
end

local function getFollowerMissionBiasRGB(missionID,followerID)
	local bias = C_Garrison.GetFollowerBiasForMission(missionID, followerID);

	if ( bias == -1 ) then -- Impossible
		return 1, 0.1, 0.1;
	elseif ( bias < 0 ) then -- Hard
		return 1, 0.5, 0.25;
	else -- Normal
		return 1, 1, 1;
	end
end

local function levelComparator(left, right)
	if (left.iLevel ~= right.iLevel ) then
		return left.iLevel > right .iLevel;
	else
		return left.level > right.level;
	end
end

local function getMissionTimeleft(follower)
	if isOnMission(follower) then
		if ( MissionMinder.ptr ) then
			if ( C_Garrison.GetFollowerMissionTimeLeftSeconds(follower.followerID) <= 0 ) then
				return L_COMPLETED, nil, nil, nil; -- Default yellow
			else
				local timeleft = C_Garrison.GetFollowerMissionTimeLeft(follower.followerID);
				return string.format("(%s)", timeleft), 0.7, 0.7, 0.7; -- Gray
			end
		else
			-- entry can be nil if follower is on a complete mission that has been viewed, but still needs to be looted.
			local entry = timeLeft[follower.followerID];
			if ( entry == nil or MissionMinder.IsMissionComplete(entry) ) then
				return L_COMPLETED, nil, nil, nil; -- Default yellow
			else
				return string.format("(%s)", timeLeft[follower.followerID].remaining), 0.7, 0.7, 0.7; -- Gray
			end
		end
	elseif (follower.status == GARRISON_FOLLOWER_WORKING) then
		return L_WORKING, 1, 0, 0; -- red
	elseif (follower.status == GARRISON_FOLLOWER_IN_PARTY) then
		return L_IN_PARTY -- use default color
	else
		return L_AVAILABLE, 0, 1, 0; -- Green
	end
end



local cache = {
	["isDirty"] = true,
	["abilities"] = {},
	["traits"] = {}
};


local function dirtyMissionCounterCache()
	cache.isDirty = true;
end

local function getFollowerCountersForMisison(followerID, missionID)
	if( cache.isDirty or not cache.abilities[missionID] or not cache.traits[missionID] ) then
		cache.abilities[missionID] = C_Garrison.GetBuffedFollowersForMission(missionID);
		cache.traits[missionID] = C_Garrison.GetFollowersTraitsForMission(missionID);
	end

	local abilities = cache.abilities[missionID];
	local traits = cache.traits[missionID];

	local hasBonusXP = MissionHasBonusXPReward(missionID);

	local counters = {};
	local hasNonFilterable = false
	if ( abilities and abilities[followerID] ) then
		for i = 1, #abilities[followerID] do
			local ability = abilities[followerID][i];
			counters[#counters + 1] = createInlineTexture(ability.icon);

			hasNonFilterable = true
		end
	end

	if ( traits and traits[followerID] ) then
		for i = 1, #traits[followerID] do
			local trait = traits[followerID][i];
			counters[#counters + 1] = createInlineTexture(trait.icon);

			if ( not ignoreTrait(trait.traitID, hasBonusXP) ) then
				hasNonFilterable = true
			end
		end
	end

	if ( hasNonFilterable ) then
		return counters;
	else
		return nil;
	end
end

local function createMissionTooltip(self, button)
	local mission = self.info;

	if ( not mission or not self:IsShown() ) then -- Not a valid button for mouseover
		return
	end

	if(mission.inProgress) then -- Do not show tooltip on inProgress missions
		return;
	end

	--[===[@debug@
	GameTooltip:AddLine("MissionID = " .. mission.missionID);
	--@end-debug@]===]
	GameTooltip:AddLine(" ");

	local party = nil;
	if ( minimizedParties[mission.missionID] ) then
		party = minimizedParties[mission.missionID]
		local hasParty = false;

		for i=1,#party do
			if ( party[i] ) then
				if (not hasParty ) then
					GameTooltip:AddLine(L_PARTY);
					hasParty = true;
				end

				local follower = C_Garrison.GetFollowerInfo(party[i]);
				local r1, g1, b1 = getFollowerMissionBiasRGB(mission.missionID, follower.followerID);
				local counters = getFollowerCountersForMisison(follower.followerID, mission.missionID);

				local level = follower.level
				if (mission.iLevel and mission.iLevel > 600 and follower.level == GARRISON_FOLLOWER_MAX_LEVEL) then
					level = follower.iLevel
				end

				local left = string.format("[%d] %s %s", level, follower.name, counters and table.concat(counters, "") or "");
				GameTooltip:AddLine(left, r1, g1, b1);
			end
		end

		if (hasParty) then
			GameTooltip:AddLine(" ");
		end
	end

	updateFollowerMissionTimeLeft(timeLeft);

	local followers = C_Garrison.GetFollowers();
	local numFollowers = followers and #followers or 0;
	table.sort(followers, levelComparator);

	local counterAvailable = false;
	for i1 = 1, numFollowers do -- Necessary to walk all followers because we want to sort by level
		local follower = followers[i1];

		if ( canBeSentOnMission(follower) and not contains(party, follower.followerID)) then
			local r1, g1, b1 = getFollowerMissionBiasRGB(mission.missionID, follower.followerID);

			local counters = getFollowerCountersForMisison(follower.followerID, mission.missionID);
			if ( counters and #counters > 0 ) then
				if ( not counterAvailable ) then
					GameTooltip:AddLine(L_COUNTERS);
					counterAvailable = true;
				end

				local level = follower.level
				if (mission.iLevel and mission.iLevel > 600 and follower.level == GARRISON_FOLLOWER_MAX_LEVEL) then
					level = follower.iLevel
				end

				local left = string.format("[%d] %s %s", level, follower.name, table.concat(counters, ""));
				local right, r2, g2, b2 = getMissionTimeleft(follower);
				GameTooltip:AddDoubleLine(left, right, r1, g1, b1, r2, g2, b2);
			end
		end
	end

	if ( not counterAvailable ) then
		GameTooltip:AddLine(L_NO_COUNTERS_AVAILABLE);
	end

	GameTooltip:Show(); -- resize the tooltip for the new lines
end

local function createFollowersTooltip(self)
	local follower = C_Garrison.GetFollowerInfo(self.id);
	if ( not follower.isCollected ) then
		return;
	end


	GameTooltip:SetOwner(self, "ANCHOR_NONE");
	GameTooltip:SetPoint("TOPRIGHT", self, "TOPLEFT", -10);
	local r, g, b = GetItemQualityColor(follower.quality);
	GameTooltip:SetText(follower.name, r, g, b);

	--[===[@debug@
	GameTooltip:AddLine("FollowerID = " .. follower.followerID);
	if ( GarrisonMissionFrame.MissionTab.MissionPage:IsShown() ) then
		local mission = GarrisonMissionFrame.MissionTab.MissionPage.missionInfo;
		local bias = C_Garrison.GetFollowerBiasForMission(mission.missionID, follower.followerID)
		GameTooltip:AddLine(string.format("Bias = %d%%", bias * 100));
	end
	--@end-debug@]===]

	if ( isOnMission(follower) ) then
		updateFollowerMissionTimeLeft(timeLeft);
		GameTooltip:AddLine(getMissionTimeleft(follower))
	end
	GameTooltip:AddLine(" ");


	local availableMissions = GarrisonMissionFrame.MissionTab.MissionList.availableMissions;
	C_Garrison.GetAvailableMissions(availableMissions);
	table.sort(availableMissions, levelComparator);

	if ( followerMissions[follower.followerID] ) then
		local mission =  followerMissions[follower.followerID]

		local r, g, b = getFollowerMissionBiasRGB(mission.missionID, follower.followerID);
		local counters = getFollowerCountersForMisison(follower.followerID, mission.missionID);
		local level = mission.level
		if (mission.iLevel and mission.iLevel > 600) then
			level = mission.iLevel
		end

		local rewards = getMissionRewards(mission);
		local left = string.format("[%d] %s (%s)", level, mission.name, rewards);
		local right = counters and table.concat(counters, "") or nil;

		GameTooltip:AddLine(L_ASSIGNED_MISSION);
		GameTooltip:AddDoubleLine(left, right, r, g, b, 1, 1, 1);
		GameTooltip:AddLine(" ");
	end

	local atleastOneMission = false;
	for _,mission in pairs(availableMissions) do
		local r, g, b = getFollowerMissionBiasRGB(mission.missionID, follower.followerID);

		local counters = getFollowerCountersForMisison(follower.followerID, mission.missionID);
		if ( counters and #counters > 0 ) then
			if( not atleastOneMission ) then
				GameTooltip:AddLine(GARRISON_MISSIONS);
				atleastOneMission = true;
			end

			local level = mission.level
			if (mission.iLevel and mission.iLevel > 600) then
				level = mission.iLevel
			end

			local rewards = getMissionRewards(mission);
			local left = string.format("[%d] %s (%s)", level, mission.name, rewards);
			local right = table.concat(counters, "");
			GameTooltip:AddDoubleLine(left, right, r, g, b, 1, 1, 1);
		end
	end

	if ( not atleastOneMission ) then
		GameTooltip:AddLine(L_NO_COUNTERS_AVAILABLE);
	end

	GameTooltip:Show();
	cache.isDirty = false;
end

local function biasColorFollowerLevels()
	local mission = GarrisonMissionFrame.MissionTab.MissionPage.missionInfo;
	local buttons = GarrisonMissionFrame.FollowerList.listScroll.buttons;

	for i = 1, #buttons do
		local button = buttons[i];
		local r, g, b = 1, 1, 1;

		if ( GarrisonMissionFrame.MissionTab.MissionPage:IsShown() and not GarrisonMissionFrame.FollowerTab:IsShown()
				and mission and button.isCollected) then
			r, g, b = getFollowerMissionBiasRGB(mission.missionID, button.id);
		end

		button.PortraitFrame.Level:SetTextColor(r, g, b)
	end
end

local function getPotentialMissionXPGains(mission, follower)
	if ( not mission or not follower ) then
		return 0
	end

	local _, base = C_Garrison.GetMissionInfo(mission.missionID);

	local bonus = 0;
	for _, reward in pairs(C_Garrison.GetMissionRewardInfo(mission.missionID)) do
		if(reward.followerXP) then
			bonus = bonus + reward.followerXP
		end
	end

	local _, _, _, _, partyBuffs, _, modify = C_Garrison.GetPartyMissionInfo(mission.missionID);
	local gain = 0;
	for i = 1, #partyBuffs do
		if ( partyBuffs[i] ~= TRAIT_FAST_LEANER ) then
			gain = gain + (expGainTraits[partyBuffs[i]] or 0);
		end
	end

	if ( follower.abilities ) then
		for i = 1, #follower.abilities do
			local ability = follower.abilities[i]
			if ( ability.isTrait and ability.id == TRAIT_FAST_LEANER ) then
				gain = gain + (expGainTraits[ability.id] or 0);
			end
		end
	end

	local bias = C_Garrison.GetFollowerBiasForMission(mission.missionID, follower.followerID);
	if ( mission.iLevel and mission.iLevel > 600 ) then
		if ( bias <= -1 ) then
			bias = ORANGE_BIAS_PERCENTAGE
		else
			bias = WHITE_BIAS_PERCENTAGE
		end
	else
		if ( bias <= -1 ) then
			bias = RED_BIAS_PERCENTAGE;
		elseif ( bias < 0) then
			bias = ORANGE_BIAS_PERCENTAGE
		else
			bias = WHITE_BIAS_PERCENTAGE;
		end
	end

	return bias * (base + modify + (bonus * (1 + gain)));
end

local function getActualMissionXPGains(potentialXP, follower)
	if( potentialXP > 0 ) then
		-- Maxed followers don't get any XP
		if ( not GarrisonMissionFrame_GetFollowerNextLevelXP(follower.level, follower.quality) ) then
			return 0;
		end

		-- Follower expected to level and max, only show limited amount of XP they would gain
		local maxXP = GarrisonMissionFrame_GetFollowerNextLevelXP(follower.level, follower.quality)
		if ( maxXP ) then
			local remainingXP = maxXP - follower.xp;
			if ( potentialXP >= remainingXP and followerWillMaxOnLevelUp(follower.level, follower.quality) ) then
				return remainingXP;
			end
		end
	end

	return potentialXP;
end

local function updateLevelUp(self)
	local missionPage = GarrisonMissionFrame.MissionTab.MissionPage
	local mission = missionPage.missionInfo;
	local follower = self.info

	if ( mission and missionPage:IsShown() and not GarrisonMissionFrame.FollowerTab:IsShown() and canBeSentOnMission(follower)) then
		local maxXP = GarrisonMissionFrame_GetFollowerNextLevelXP(follower.level, follower.quality);
		local missionXPGains = getPotentialMissionXPGains(mission, follower);

		if ( maxXP and follower.xp + missionXPGains >= maxXP ) then
			self.levelUp:SetAtlas(LEVELUP_ARROW_QUALITIES[follower.quality], true);
			self.levelUp:Show();
		else
			self.levelUp:Hide();
		end
	else
		self.levelUp:Hide();
	end
end

local function updateXPGains(self)
	local missionPage = GarrisonMissionFrame.MissionTab.MissionPage
	local mission = missionPage.missionInfo;
	local follower = self.info

	local potentialXP = getPotentialMissionXPGains(mission, follower)
	local missionXPGains = getActualMissionXPGains(potentialXP, follower);
	if (missionXPGains > 0) then
		self.XPGain:SetFormattedText(XP_GAIN, BreakUpLargeNumbers(math.ceil(missionXPGains)));
		self.XPGain:Show();
	else
		self.XPGain:Hide();
	end
end

local function updateXpRemaining(self)
	local follower = self.info

	if ( follower and follower.isCollected ) then
		local maxXP = GarrisonMissionFrame_GetFollowerNextLevelXP(follower.level, follower.quality);
		if ( maxXP ) then
			self.xpRemaining:SetFormattedText(GARRISON_REWARD_XP_FORMAT, BreakUpLargeNumbers(maxXP - follower.xp));
		else
			self.xpRemaining:SetText(nil);
		end
	else
		self.xpRemaining:SetText(nil);
	end
end

local function updateFollowerLevelUp()
	local buttons = GarrisonMissionFrame.FollowerList.listScroll.buttons;
	for i = 1, #buttons do
		updateLevelUp(buttons[i]);
	end
end

local function updateFollowerXPRemaining()
	local buttons = GarrisonMissionFrame.FollowerList.listScroll.buttons;
	for i = 1, #buttons do
		updateXpRemaining(buttons[i]);
	end
end

local function updatePartyLevelUp()
	local partyFollowers = GarrisonMissionFrame.MissionTab.MissionPage.Followers;
	for i = 1, #partyFollowers do
		updateLevelUp(partyFollowers[i]);
	end
end

local function updatePartyXPGains()
	local partyFollowers = GarrisonMissionFrame.MissionTab.MissionPage.Followers;
	for i = 1, #partyFollowers do
		updateXPGains(partyFollowers[i]);
	end
end

local function hideMissionCountersTooltip(self)
	GameTooltip:Hide();
end


local singleclickTimer = nil;
-- Override the standard OnClick event to delay just long enough to determine if it is a DoubleClick.
--If it is a double click then the OnClick event will not fire. If it isn't a double click then the
--OnClick event (which will run via timer) will still run; albeit slightly delayed.
local function followerSingleClick(self, button)
	if(singleclickTimer) then
		singleclickTimer:Cancel();
	end

	-- Only delay OnClick in very specific cases
	if ( self.isCollected and button == "LeftButton" ) then
		if ( GarrisonMissionFrame.MissionTab.MissionPage:IsShown() and not GarrisonMissionFrame.FollowerTab:IsShown() ) then
			singleclickTimer = C_Timer.NewTimer(0.2, function()
				module.GarrisonFollowerListButton_OnClick(self, button);
			end);
			return;
		end
	end

	module.GarrisonFollowerListButton_OnClick(self, button);
end

-- Handle OnDoubleClick event and disable the single click timer
local function assignFollowerToMission(self, button)
	if(singleclickTimer) then
		singleclickTimer:Cancel();
	end

	-- Hide the rightclick menu
	GarrisonFollowerOptionDropDown.followerID = nil;
	CloseDropDownMenus();

	if ( self.isCollected and self.info.status == nil and button == "LeftButton" ) then
		local mission = GarrisonMissionFrame.MissionTab.MissionPage.missionInfo;
		if ( mission and GarrisonMissionFrame.MissionTab.MissionPage:IsShown()
					and not GarrisonMissionFrame.FollowerTab:IsShown() ) then
			DEBUG("assignFollowerToMission(mission=%s, follower=%s)", mission.missionID, self.id);
			GarrisonMissionPage_AddFollower(self.id);
		end
	end
end

local function disableTimer()
	if(singleclickTimer) then
		singleclickTimer:Cancel()
		singleclickTimer = nil
	end
end

local followerDragSource = nil;
local function setDragSource(self, button)
	if (GarrisonFollowerPlacer:IsShown()) then
		DEBUG("SetDragSource(%s)", tostring(self));
		followerDragSource = self;
	end
end

local function clearDragSource()
DEBUG("ClearDragSource()");
	followerDragSource = nil;
end

local function getFollowerFrameIndex(self)
	local partyFollowers = GarrisonMissionFrame.MissionTab.MissionPage.Followers;
	for i = 1, #partyFollowers do
		if ( partyFollowers[i] == self) then
			return i
		end
	end

	return 0;
end

local function setFollower(destFrame, destInfo, ...)
	DEBUG("SetFollower(%s, %s) DragSource=%s", tostring(destFrame), destInfo.followerID, tostring(followerDragSource));
	if ( followerDragSource ) then
		local srcFrame = followerDragSource;
		local srcInfo = destFrame.info
		srcIndex = getFollowerFrameIndex(srcFrame);
		destIndex = getFollowerFrameIndex(destFrame);

		DEBUG("srcIndex=%d, destIndex=%d", srcIndex, destIndex)

		if (srcIndex > destIndex ) then
			-- We always want to assign the "lower" frame first to keep the follower order consistent
			--There are some internals to garrison missions that can cause weirdness for the user.
			--like hidden enemy modifiers on [615] Auchindoom!
			srcFrame, destFrame = destFrame, srcFrame
			srcInfo, destInfo = destInfo, srcInfo
		elseif (srcIndex < destIndex ) then
			-- Must first unassign dest follower
			GarrisonMissionPage_ClearFollower(destFrame);
		else -- Targetting the same frame, only do one set (destination)
			srcInfo = nil;
		end

		if(srcInfo) then
			DEBUG("src SetFollower(%s, %s)", tostring(srcFrame), srcInfo.followerID);
			module.GarrisonMissionPage_SetFollower(srcFrame, srcInfo, ...)
		end
		if (destInfo) then
			DEBUG("dest SetFollower(%s, %s)", tostring(destFrame), destInfo.followerID);
			module.GarrisonMissionPage_SetFollower(destFrame, destInfo, ...)
		end
	else
		module.GarrisonMissionPage_SetFollower(destFrame, destInfo, ...)
	end
end

local function ShowBonusXPAmount(self, rewards, numRewards)
	if (numRewards > 0 and rewards) then
		local index = 1;
		for id, reward in pairs(rewards) do
			if (self.Rewards[index]) then -- Should never be nil, but if it ever is we ignore it
				local Reward = self.Rewards[index];
				if ( reward.followerXP and reward.followerXP > 0 ) then
					Reward.Quantity:SetText(BreakUpLargeNumbers(reward.followerXP));
					Reward.Quantity:Show();
				elseif (reward.currencyID == 0) then -- gold
					Reward.Quantity:SetText(BreakUpLargeNumbers(math.floor(reward.quantity / COPPER_PER_GOLD)));
					Reward.Quantity:Show();
				end
			end

			index = index + 1;
		end
	end
end

local function updateRewardQualityBorder(frame, reward)
	if ( not frame.IconBorder ) then
		local iconborder = frame:CreateTexture("IconBorder", "OVERLAY")
		iconborder:SetTexture("Interface\\Common\\WhiteIconFrame")
		iconborder:SetAllPoints(frame.Icon)
		frame.IconBorder = iconborder
	end

	local itemID = reward.itemID
	if ( itemID ) then
		local _, _, quality = GetItemInfo(itemID);
		if  ( quality and quality > LE_ITEM_QUALITY_COMMON and BAG_ITEM_QUALITY_COLORS[quality]) then
			frame.IconBorder:Show();
			frame.IconBorder:SetVertexColor(BAG_ITEM_QUALITY_COLORS[quality].r, BAG_ITEM_QUALITY_COLORS[quality].g, BAG_ITEM_QUALITY_COLORS[quality].b);
		else
			frame.IconBorder:Hide();
		end
	else
		frame.IconBorder:Hide();
	end
end

local function MissionButton_updateRewardQualityBorder(self, rewards, numRewards)
	if (numRewards > 0 and rewards) then
		local index = 1;
		for id, reward in pairs(rewards) do
			local frame = self.Rewards[index];
			if ( frame ) then -- Should never be nil, but if it ever is we ignore it
				updateRewardQualityBorder(frame, reward)
			end
			index = index + 1;
		end
	end
end

local function LandingMissionButton_updateRewardQualityBorder()
	local missions = GarrisonLandingPageReport.List.AvailableItems or {};
	local buttons = GarrisonLandingPageReport.List.listScroll.buttons;

	for i = 1, #buttons do
		local button = buttons[i]
		local mission = missions[button.id]
		if ( mission ) then
			local rewards = mission.rewards
			local numRewards = mission.numRewards

			MissionButton_updateRewardQualityBorder(button, rewards, numRewards)
		end
	end
end


local function clearAllParties()
	minimizedParties = {}
	local missions = C_Garrison.GetAvailableMissions();
	for i = 1, #missions do
		local mission = missions[i];
		for i2 = 1, #mission.followers do
			local followerID = mission.followers[i2];
			C_Garrison.RemoveFollowerFromMission(mission.missionID, followerID);
		end
	end
end

function setFollowerFromFrame(frame, followerID)
	local missionPage = GarrisonMissionFrame.MissionTab.MissionPage
	if ( not followerID ) then
		GarrisonMissionPage_ClearFollower(frame);
		return;
	elseif (frame.info) then
		GarrisonMissionPage_ClearFollower(frame);
	end

	local info = C_Garrison.GetFollowerInfo(followerID);
	frame.info = info;
	frame.Name:Show();
	frame.Name:SetText(info.name);
	if (frame.Class) then
		frame.Class:Show();
		frame.Class:SetAtlas(info.classAtlas);
	end
	frame.PortraitFrame.Empty:Hide();

	local showItemLevel;
	if ( missionPage.showItemLevel and info.level == GarrisonMissionFrame.followerMaxLevel ) then
		frame.PortraitFrame.LevelBorder:SetAtlas("GarrMission_PortraitRing_iLvlBorder");
		frame.PortraitFrame.LevelBorder:SetWidth(70);
		showItemLevel = true;
	else
		frame.PortraitFrame.LevelBorder:SetAtlas("GarrMission_PortraitRing_LevelBorder");
		frame.PortraitFrame.LevelBorder:SetWidth(58);
		showItemLevel = false;
	end
	GarrisonMissionFrame_SetFollowerPortrait(frame.PortraitFrame, info, showItemLevel);

	local counters = GarrisonMissionFrame.followerCounters and GarrisonMissionFrame.followerCounters[frame.info.followerID] or nil;
	if (counters) then
		for i = 1, #counters do
			if (not frame.Counters[i]) then
				frame.Counters[i] = CreateFrame("Frame", nil, frame, "GarrisonMissionAbilityLargeCounterTemplate");
				frame.Counters[i]:SetPoint("LEFT", frame.Counters[i-1], "RIGHT", 16, 0);
			end
			local Counter = frame.Counters[i];
			Counter.info = counters[i];
			Counter.info.showCounters = true;
			Counter.Icon:SetTexture(counters[i].icon);
			Counter.tooltip = counters[i].name;
			Counter:Show();
		end
		for i = (#counters + 1), #frame.Counters do
			frame.Counters[i]:Hide();
		end
	end

	-- update follower list
	GarrisonMissionFrame.followerTraits = C_Garrison.GetFollowersTraitsForMission(missionPage.missionInfo.missionID);
	GarrisonFollowerList_UpdateFollowers(GarrisonMissionFrame.FollowerList);

	local followerBias = C_Garrison.GetFollowerBiasForMission(missionPage.missionInfo.missionID, info.followerID);
	if ( followerBias == -1 ) then
		frame.PortraitFrame.Level:SetTextColor(1, 0.1, 0.1);
	elseif ( followerBias < 0 ) then
		frame.PortraitFrame.Level:SetTextColor(1, 0.5, 0.25);
	else
		frame.PortraitFrame.Level:SetTextColor(1, 1, 1);
	end
	GarrisonMissionPage_UpdateMissionForParty();

	if ( missionPage.missionInfo.numFollowers == 1 ) then
		local model = missionPage.FollowerModel;
		model:Show();
		model:SetTargetDistance(0);
		GarrisonMission_SetFollowerModel(model, info.followerID, info.displayID);
		model:SetHeightFactor(info.height or 1);
		model:InitializeCamera(info.scale or 1);
		model:SetFacing(-.2);
		missionPage.EmptyFollowerModel:Hide();
	end

	GarrisonMissionPage_SetCounters();
end

local function clearFollowerFromFrame(frame)
	local missionPage = GarrisonMissionFrame.MissionTab.MissionPage
	local followerID = frame.info and frame.info.followerID or nil;
	frame.info = nil;
	frame.Name:Hide();
	if (frame.Class) then
		frame.Class:Hide();
	end
	frame.PortraitFrame.Empty:Show();
	frame.PortraitFrame.LevelBorder:SetAtlas("GarrMission_PortraitRing_LevelBorder");
	frame.PortraitFrame.LevelBorder:SetWidth(58);
	frame.PortraitFrame.Level:SetText("");
	for i = 1, #frame.Counters do
		frame.Counters[i]:Hide();
	end

	if (followerID) then
		if ( missionPage.missionInfo.numFollowers == 1 ) then
			missionPage.FollowerModel:ClearModel();
			missionPage.FollowerModel:Hide();
			missionPage.EmptyFollowerModel:Show();
		end
	end

	GarrisonMissionPage_SetCounters();

	return followerID
end

local function loadSavedParty(self, ...)
	local missionPage = GarrisonMissionFrame.MissionTab.MissionPage
	local mission = self.info

	if ( mission and minimizedParties[mission.missionID] ) then
		local followers = missionPage.Followers;
		local party = minimizedParties[mission.missionID];
		for i=1,#followers do
			local followerID = party[i];
			DEBUG("Restoring follower [%d] = %s", i, tostring(followerID));
			setFollowerFromFrame(followers[i], followerID);
			party[i] = nil;
			if ( followerID ) then
				followerMissions[followerID] = nil
			end
		end
		updatePartyLevelUp()
		updatePartyXPGains()
	end
end

function MissionMinder_MinimizeMission(self, ...)
	local missionPage = GarrisonMissionFrame.MissionTab.MissionPage
	local mission = missionPage.missionInfo
	if( not mission ) then
		return
	end

	if ( not minimizedParties[mission.missionID] ) then
		minimizedParties[mission.missionID] = {}
	end
	local party = minimizedParties[mission.missionID]

	for i = 1, #missionPage.Followers do
		local followerFrame = missionPage.Followers[i];
		local followerID = clearFollowerFromFrame(followerFrame);
		party[i] = followerID;
		if ( followerID ) then
			followerMissions[followerID] = mission
		end

		DEBUG("Saving follower [%d][%d] = %s", mission.missionID, i, tostring(party[i]));
	end
	updatePartyLevelUp();
	updatePartyXPGains();
	missionPage.FollowerModel:Hide();
	GarrisonMissionPage_UpdateEmptyString();

	GarrisonMissionFrame.MissionTab.MissionPage:Hide();
	GarrisonMissionFrame.MissionTab.MissionList:Show();
	GarrisonMissionFrame.followerCounters = nil;
	GarrisonMissionFrame.followerTraits = nil;
	GarrisonMissionFrame.MissionTab.MissionPage.missionInfo = nil;
end

local MINUTES = 60;
local HOURS = 60 * MINUTES;
local DAYS = 24 * HOURS;

local function getCompactDurationString(time)
	local str;
	if (time >= DAYS) then
		time = ((time - 1) / DAYS) + 1;
		return string.format(COOLDOWN_DURATION_DAYS, time);
	elseif (time >= HOURS) then
		time = ((time - 1) / HOURS) + 1;
		return string.format(COOLDOWN_DURATION_HOURS, time);
	elseif (time >= MINUTES) then
		time = ((time - 1) / MINUTES) + 1;
		return string.format(COOLDOWN_DURATION_MIN, time);
	else
		return string.format(COOLDOWN_DURATION_SEC, time);
	end
end

local function showMissionAvailabilityTimeRemaining()
	local self = GarrisonMissionFrame.MissionTab.MissionList;

	for _,button in pairs(self.listScroll.buttons) do
		local mission = button.info

		if ( mission and mission.offerEndTime and not self.showInProgress ) then
			local timeRemaining = mission.offerEndTime - GetTime();
			if ( timeRemaining > 0 ) then
				button.Expires:SetText(getCompactDurationString(timeRemaining))
				if ( timeRemaining < 2*HOURS ) then
					button.Expires:SetTextColor(1, 0, 0);
				else
					button.Expires:SetTextColor(1, 1, 1);
				end
				button.Expires:Show();
			else
				button.Expires:Hide();
			end
		else
			button.Expires:Hide()
		end
	end
end

local function updateMissionForMinimizedParties()
	local self = GarrisonMissionFrame.MissionTab.MissionList;

	if (self.showInProgress) then
		return
	end

	for _,button in pairs(self.listScroll.buttons) do
		local mission = button.info

		if ( mission ) then
			local _, _, _, successChance = C_Garrison.GetPartyMissionInfo(mission.missionID);
			local party = minimizedParties[mission.missionID]

			if ( not isEmpty(party) and successChance ) then
				MissionMinder_UpdatePartyMembers(button, party, mission.numFollowers)
				button.SuccessChance:SetFormattedText(PERCENTAGE_STRING .. " " .. GARRISON_MISSION_CHANCE, successChance);
				button.SuccessChance:Show();
			else
				MissionMinder_UpdatePartyMembers(button, nil)
				button.SuccessChance:SetText(nil);
				button.SuccessChance:Hide();
			end
		end
	end
end

local function determineCountersFromMinimizedParty(missionID)
	local threats = {};
	threats.full = {};
	threats.partial = {};
	threats.away = {};
	threats.worker = {};


	if ( not minimizedParties[missionID] ) then
		minimizedParties[missionID] = {}
	end
	local party = minimizedParties[missionID]

	for i = 1, #party do
		local followerID = party[i];
		if ( followerID ) then
			local follower = C_Garrison.GetFollowerInfo(followerID);

			if ( follower.isCollected and follower.status ~= GARRISON_FOLLOWER_INACTIVE ) then
				local bias = C_Garrison.GetFollowerBiasForMission(missionID, followerID);
				if ( bias > -1.0 ) then
					local abilities = C_Garrison.GetFollowerAbilities(followerID);
					for j = 1, #abilities do
						for counterMechanicID in pairs(abilities[j].counters) do
							local isFullCounter = C_Garrison.IsMechanicFullyCountered(missionID, followerID, counterMechanicID, abilities[j].id);
							local map = ( isFullCounter ) and threats.full or threats.partial;
							map[counterMechanicID] = (map[counterMechanicID] or 0) + 1
						end
					end
				end
			end
		end
	end

	return threats;
end

local nextFollowerListUpdate = nil;
local function followerList_OnUpdate(self)
	local now = GetTime();
	if ( not nextFollowerListUpdate or now > nextFollowerListUpdate ) then
		nextMissionListUpdate = math.floor(now + 1)
		GarrisonFollowerList_Update(GarrisonMissionFrame)
	end
end

local function initialize()
--function module.initialize()
	DEBUG("   |cffffffffPlanner|r initialized")

	-- Mission tooltips
	local missionList = GarrisonMissionFrame.MissionTab.MissionList
	local missionButtons = missionList.listScroll.buttons;
	for i = 1, #missionButtons do
		local button = missionButtons[i]
		button:HookScript("OnEnter", createMissionTooltip);
	end
	if ( MissionMinder.ptr ) then
		hooksecurefunc("GarrisonMissionButton_OnEnter", createMissionTooltip)
	end

	-- Update mission tooltip mechanics with minimized party counters
	if ( MissionMinder.ptr ) then
		GarrisonMission_DetermineCounterableThreats = determineCountersFromMinimizedParty
	end

	-- Load minimized mission party
	for i = 1, #missionButtons do
		local button = missionButtons[i]
		button:HookScript("OnClick", loadSavedParty);
	end

	-- Show mission expiration
	if ( MissionMinder.ptr ) then
		for i = 1, #missionButtons do
			local button = missionButtons[i]
			button.Expires = button:CreateFontString("MissionExpires", "ARTWORK", "GameTooltipText")
			button.Expires:SetPoint("BOTTOM", button.MissionType);
			button.Expires:Show();
		end
		hooksecurefunc("GarrisonMissionList_Update", showMissionAvailabilityTimeRemaining)
		missionList.listScroll.update = GarrisonMissionList_Update;  -- Set the update method to the new hooked method
	end

	-- Show minimized parties on mission button
	hooksecurefunc("GarrisonMissionList_Update", updateMissionForMinimizedParties)
	missionList.listScroll.update = GarrisonMissionList_Update;  -- Set the update method to the new hooked method

	-- Follower tooltips
	local FollowerList = GarrisonMissionFrame.FollowerList
	local followerButtons = FollowerList.listScroll.buttons;
	for i = 1, #followerButtons do
		local button = followerButtons[i]
		button:SetScript("OnEnter", createFollowersTooltip);
		button:SetScript("OnLeave", hideMissionCountersTooltip);
	end

	-- Double click to assign followers to mission
	module.GarrisonFollowerListButton_OnClick = GarrisonFollowerListButton_OnClick;
	GarrisonMissionFrame.MissionTab.MissionPage:HookScript("OnHide", disableTimer);
	for i = 1, #followerButtons do
		local button = followerButtons[i]
		button:SetScript("OnClick", followerSingleClick);
		button:SetScript("OnDoubleClick", assignFollowerToMission);
	end

	-- Show remaining XP to level on follower list button
	hooksecurefunc("GarrisonFollowerList_Update", updateFollowerXPRemaining)
	for i = 1, #followerButtons do
		local button = followerButtons[i]
		local xpRemaining = button:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall");
		xpRemaining:SetPoint("BOTTOMRIGHT", button.BG);
		xpRemaining:SetText(nil);
		xpRemaining:Show();
		button.xpRemaining = xpRemaining;
	end

	-- Color follower levels by bias
	hooksecurefunc("GarrisonFollowerList_Update", biasColorFollowerLevels)

	-- Drag-to-swap party members
	module.GarrisonMissionPage_SetFollower = GarrisonMissionPage_SetFollower
	GarrisonMissionPage_SetFollower = setFollower
	GarrisonFollowerPlacerFrame:HookScript("OnClick", clearDragSource);
	local partyFollowers = GarrisonMissionFrame.MissionTab.MissionPage.Followers;
	for i = 1, #partyFollowers do
		partyFollowers[i]:HookScript("OnDragStart", setDragSource);
		partyFollowers[i]:HookScript("OnReceiveDrag", clearDragSource);
	end
	for i = 1, #followerButtons do
		local button = followerButtons[i]
		button:HookScript("OnDragStart", clearDragSource)
	end

	-- Follower level up
	hooksecurefunc("GarrisonFollowerList_Update", updateFollowerLevelUp)
	hooksecurefunc("GarrisonMissionPage_SetFollower", updatePartyLevelUp);
	hooksecurefunc("GarrisonMissionPage_ClearFollower", updatePartyLevelUp);
	for i = 1, #followerButtons do
		local button = followerButtons[i]
		local levelUp = button.PortraitFrame:CreateTexture();
		levelUp:SetPoint("BOTTOM", button.PortraitFrame, "BOTTOMLEFT")
		levelUp:Hide();
		button.levelUp = levelUp
	end
	for i = 1, #partyFollowers do
		local levelUp = partyFollowers[i].PortraitFrame:CreateTexture();
		levelUp:SetPoint("BOTTOM", partyFollowers[i].PortraitFrame, "BOTTOMLEFT")
		levelUp:Hide();
		partyFollowers[i].levelUp = levelUp
	end

	-- Estimate party member XP gains
	hooksecurefunc("GarrisonMissionPage_SetFollower", updatePartyXPGains);
	hooksecurefunc("GarrisonMissionPage_ClearFollower", updatePartyXPGains);
	for i = 1, #partyFollowers do
		local xpGain = partyFollowers[i]:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
		xpGain:SetPoint("TOP", partyFollowers[i], "BOTTOM")
		xpGain:SetTextColor(0, 1, 0)
		xpGain:SetText(nil)
		xpGain:Show()
		partyFollowers[i].XPGain = xpGain
	end

	-- Periodically update follower list (for mission time remaining)
	if ( MissionMinder.ptr ) then
		FollowerList:SetScript("OnUpdate", followerList_OnUpdate);
	end

	-- Show reward amounts
	if ( not MissionMinder.ptr ) then
		hooksecurefunc("GarrisonMissionButton_SetRewards", ShowBonusXPAmount);
	end

	-- Show reward quality borders
	hooksecurefunc("GarrisonMissionButton_SetRewards", MissionButton_updateRewardQualityBorder);
	hooksecurefunc("GarrisonLandingPageReportList_UpdateAvailable", LandingMissionButton_updateRewardQualityBorder);
	hooksecurefunc("GarrisonMissionPage_SetReward", updateRewardQualityBorder);

	-- Clear all parties when mission frame closes
	GarrisonMissionFrame:HookScript("OnHide", clearAllParties)
end


local events = Lib.events;
events:RegisterEvent("planner", "ADDON_LOADED", initialize);
events:RegisterEvent("planner", "GARRISON_FOLLOWER_LIST_UPDATE", dirtyMissionCounterCache);
events:RegisterEvent("planner", "GARRISON_FOLLOWER_XP_CHANGED", dirtyMissionCounterCache);
events:RegisterEvent("planner", "GARRISON_FOLLOWER_REMOVED", dirtyMissionCounterCache);
