local Lib = MissionMinder;
local PreLoadLib = MissionMinderPreLoad;
Lib.callbacks = Lib.callbacks or LibStub("CallbackHandler-1.0"):New(Lib)

local options = Lib.options;
local DEBUG = Lib.debug;

Lib.tracker = {};
local module = Lib.tracker;

local L = LibStub("AceLocale-3.0"):GetLocale("MissionMinder");
local L_NO_DATA_AVAILABLE = L["No data available"];
local L_LOADING = LFG_LIST_LOADING

local SECONDS = 1
local MINUTES = 60 * SECONDS
local HOURS = 60 * MINUTES

local CURRENT_PLAYER = GetUnitName("player") .. "-" .. GetRealmName()

local selectedPlayer = CURRENT_PLAYER
local inprogressMissions = nil
local nextSortAt = nil


local function isCurrentPlayerSelected()
	return CURRENT_PLAYER == selectedPlayer
end

local function orderInProgressItems(item1, item2)
	local now = time()

	if ( item1.isBuilding ~= item2.isBuilding ) then
		return item1.isBuilding
	end

	local item1Complete = now >= item1.endDate
	local item2Complete = now >= item2.endDate
	if ( item1Complete ~= item2Complete ) then
		return item1Complete
	end

	if ( item1.iLevel ~= item2.iLevel ) then
		return item1.iLevel > item2.iLevel
	end

	if ( item1.level ~= item2.level ) then
		return item1.level > item2.level
	end

	return strcmputf8i(item1.name, item2.name) < 0;
end

local function maybeSortInProgressMissionsForSelectedCharacter(tbl, now)
	if ( nextSortAt and now < nextSortAt ) then
		return
	end

	table.sort(tbl, orderInProgressItems)
	nextSortAt = time() + (1 * 60)
	for i = 1, #tbl do
		if ( now < tbl[i].endDate ) then
			nextSortAt = math.min(nextSortAt, tbl[i].endDate);
		end
	end
end

local function loadInProgressItems()
	if ( isCurrentPlayerSelected() ) then
		return nil
	end

	local tbl = {};
	local buildings = PreLoadLib.Globals.Characters[selectedPlayer].buildings
	if ( buildings ) then
		for _, building in pairs(buildings) do
			if ( building.isBuilding or building.canActivate ) then
				tinsert(tbl, building)
			end
		end
	end

	local missions = PreLoadLib.Globals.Characters[selectedPlayer].missions
	if ( missions ) then
		for _, mission in pairs(missions) do
			tinsert(tbl, mission);
		end
	end

	nextSortAt = nil;
	maybeSortInProgressMissionsForSelectedCharacter(tbl, time())

	return tbl
end

local function selectCharacter(player)
	DEBUG("selectedPlayer = %s", player)

	selectedPlayer = player
	GarrisonLandingPageReport.CharacterName:SetText(selectedPlayer)

	local self = GarrisonLandingPageReport
	self.EmptyString:SetText(L_LOADING);
	self.EmptyString:Show();
	for i = 1, #self.Shipments do
		self.Shipments[i]:Hide();
	end

	inprogressMissions = loadInProgressItems()

	if ( GarrisonLandingPage:IsShown() ) then
		C_Garrison.RequestLandingPageShipmentInfo();
		GarrisonLandingPageReportList_UpdateItems();
	end

	if ( GarrisonLandingPageReport.selectedTab == GarrisonLandingPageReport.Available ) then
		GarrisonLandingPageReport_SetTab(GarrisonLandingPageReport.InProgress);
	end

	GarrisonLandingPageReport.Available:SetEnabled( isCurrentPlayerSelected() )

	-- Hide button status to reset the interface (blizzard ui doesn't natively do this)
	local buttons = GarrisonLandingPageReport.List.listScroll.buttons;
	for i = 1, #buttons do
		buttons[i].Status:SetText(nil)
		buttons[i].Status:Hide()
	end

	--[===[@debug@
	for i = 1, #buttons do
		buttons[i].Seconds:SetText(nil)
		buttons[i].Seconds:Hide()
	end
	--@end-debug@]===]
end

local function selectCurrentPlayer()
	selectCharacter(CURRENT_PLAYER)
end

local function selectCharacterFromDropdown(self)
	if ( UIDropDownMenu_GetCurrentDropDown() == MissionMinderCharacterMenu ) then
		CloseDropDownMenus();
	end

	if ( self.value == selectedPlayer ) then
		return
	end

	selectCharacter(self.value)
end

function MissionMinderGarrisonResourceCache_OnEnter(self)
		GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
		GameTooltip:SetText(GARRISON_CACHE);
		GameTooltip:Show()
end

local function trackMission(event, missionID, ...)
	local now = time()

	local mission = nil

	-- Find this mission and save its details
	GarrisonMissionList_UpdateMissions()
	local missions = GarrisonMissionFrame.MissionTab.MissionList.availableMissions --C_Garrison.GetLandingPageItems()
	for i=1, #missions do
		if ( missionID == missions[i].missionID ) then
			mission = {}
			mission.missionID = missions[i].missionID
			mission.level = missions[i].level
			mission.iLevel = missions[i].iLevel
			mission.name = missions[i].name
			mission.type = missions[i].type
			mission.typeAtlas = missions[i].typeAtlas
			mission.location = missions[i].location
			mission.locPrefix = missions[i].locPrefix
			mission.isRare = missions[i].isRare

			mission.numFollowers = missions[i].numFollowers
			mission.followers = {}
			for ii = 1, #missions[i].followers do
				local followerID = missions[i].followers[ii]
				tinsert(mission.followers, C_Garrison.GetFollowerName(followerID))
			end

			mission.numRewards = missions[i].numRewards
			mission.rewards = {}
			for k,reward in pairs(missions[i].rewards) do
				if ( reward.quality ) then
					tinsert(mission.rewards, ITEM_QUALITY_COLORS[reward.quality + 1].hex..reward.title..FONT_COLOR_CODE_CLOSE)
				elseif ( reward.itemID ) then
					local itemName, _, itemRarity = GetItemInfo(reward.itemID);
					tinsert(mission.rewards, ITEM_QUALITY_COLORS[itemRarity].hex..itemName..FONT_COLOR_CODE_CLOSE);
				else
					tinsert(mission.rewards, reward.title)
				end
			end

			break;
		end
	end

	local _, totalTimeInSeconds = C_Garrison.GetPartyMissionInfo(missionID);
	if (mission and totalTimeInSeconds) then
		mission.endDate = now + totalTimeInSeconds
	end


	-- Save the mission to globals
	missions = PreLoadLib.Globals.Characters[CURRENT_PLAYER].missions
	if ( not missions ) then
		missions = {}
		PreLoadLib.Globals.Characters[CURRENT_PLAYER].missions = missions
	end

	missions[missionID] = mission
end

local function untrackMission(event, missionID)
	local missions = PreLoadLib.Globals.Characters[CURRENT_PLAYER].missions
	if ( missions ) then
		missions[missionID] = nil
	end
end

local function trackFinishedMissions()
	local trackedMissions = PreLoadLib.Globals.Characters[CURRENT_PLAYER].missions or {}
	local completedMissions = GarrisonMissionFrame.MissionComplete.completeMissions or {}

	local now = time()
	for i = 1, #completedMissions do
		local mission = completedMissions[i]
		local tracked = trackedMissions[mission.missionID]
		if ( tracked and now < tracked.endDate) then
			DEBUG("trackFinishedMissions() - Setting mission %d \"%s\" to complete", mission.missionID, mission.name)
			tracked.endDate = now - 1
		end
	end
end

local function cleanupOrphanedMissions(self)
	local trackedMissions = PreLoadLib.Globals.Characters[CURRENT_PLAYER].missions
	local inprogress = GarrisonMissionFrame.MissionTab.MissionList.inProgressMissions

	if ( inprogress and trackedMissions) then
		-- Create map of tracked missions
		local map = {}
		for missionID, mission in pairs(trackedMissions) do
			map[missionID] = true;
		end

		-- Remove missions that are still in progress (leaving only "orphaned" misisons))
		for i = 1, #inprogress do
			local mission = inprogress[i];
			if ( map[mission.missionID] ) then
				map[mission.missionID] = nil;
			else
				-- Unknown mission, in the future might be abel to take unknown missions and track
				--them (need some way to know the enddate without parsing time strings)
			end
		end

		-- Untrack orphaned missions
		for missionID, _ in pairs(map) do
			local mission = trackedMissions[missionID]
			DEBUG("Untracking mission \"%s\" (%d) because it was completed through other means", mission.name, mission.missionID)
			trackedMissions[missionID] = nil
		end
	end
end

local function isChecked(self)
	return self.value == selectedPlayer
end

function MissionMinderChangeMenu_Toggle()
	ToggleDropDownMenu(1, nil, GarrisonLandingPageReport.CharacterMenu, GarrisonLandingPageReport.ChangeCharacter, 0, 0);
end

function MissionMinderCharacterMenuDropDown_Initialize()
	local info = UIDropDownMenu_CreateInfo();
	info.isNotRadio = true;

	local characters = {}
	for character , _ in pairs(PreLoadLib.Globals.Characters) do
		tinsert(characters, character)
	end

	sort(characters, function(character1, character2)
		if ( character1 == CURRENT_PLAYER ) then
			return true
		elseif ( character2 == CURRENT_PLAYER ) then
			return false
		else
			return strcmputf8i(character1, character2) < 0
		end
	end);

	for i = 1, #characters do
		info.text = characters[i]
		info.value = characters[i]
		info.checked = isChecked
		info.func = selectCharacterFromDropdown
		UIDropDownMenu_AddButton(info)
	end
end

local function updateMissionTooltipForSelectedCharacter(self, button, ...)
	if ( isCurrentPlayerSelected() ) then
		module.GarrisonLandingPageReportMission_OnEnter(self, button, ...)
		return
	end

	if ( not self.id or not self:IsShown() ) then -- Button is not valid for mouseover
		return
	end

	local now = time()

	GameTooltip:SetOwner(self, "ANCHOR_CURSOR_RIGHT")
	local items = inprogressMissions or {}
	local item = items[self.id]

	GameTooltip:SetText(item.name)

	if ( item.isBuilding or item.canActivate ) then
		GameTooltip:AddLine(string.format(GARRISON_BUILDING_LEVEL_LABEL_TOOLTIP, item.rank), 1, 1, 1)
		if(now >= item.endDate) then
			GameTooltip:AddLine("(" .. COMPLETE .. ")", 0, 1, 0)
		else
			GameTooltip:AddLine(PreLoadLib.tracker.secondsToString(item.endDate - now).." "..RED_FONT_COLOR_CODE..GARRISON_MISSION_IN_PROGRESS..FONT_COLOR_CODE_CLOSE, 1, 1, 1)
		end

		--[===[@debug@
		GameTooltip:AddLine("BuildingID = " .. item.buildingID)
		--@end-debug@]===]

		GameTooltip:Show()
		return;
	end

	-- level
	if ( Lib.ptr ) then
		if ( item.level == GARRISON_FOLLOWER_MAX_LEVEL and item.iLevel > 0 ) then
			GameTooltip:AddLine(format(GARRISON_MISSION_LEVEL_ITEMLEVEL_TOOLTIP, item.level, item.iLevel), 1, 1, 1)
		else
			GameTooltip:AddLine(format(GARRISON_MISSION_LEVEL_TOOLTIP, item.level), 1, 1, 1)
		end
	end

	-- completed?
	if(now >= item.endDate) then
		GameTooltip:AddLine("(" .. COMPLETE .. ")", 0, 1, 0)
	else
		GameTooltip:AddLine(PreLoadLib.tracker.secondsToString(item.endDate - now).." "..RED_FONT_COLOR_CODE..GARRISON_MISSION_IN_PROGRESS..FONT_COLOR_CODE_CLOSE, 1, 1, 1)
	end

	-- success chance
	if ( item.successChance ) then
		GameTooltip:AddLine(format(GARRISON_MISSION_PERCENT_CHANCE, item.successChance), 1, 1, 1)
	end

	--[===[@debug@
	GameTooltip:AddLine("MissionID = " .. item.missionID)
	--@end-debug@]===]

	if ( item.rewards ) then
		GameTooltip:AddLine(" ")
		GameTooltip:AddLine(REWARDS)
		for i = 1, #item.rewards do
			GameTooltip:AddLine(item.rewards[i], 1, 1, 1)
		end
	end

	if ( item.followers ) then
		GameTooltip:AddLine(" ")
		GameTooltip:AddLine(GARRISON_FOLLOWERS)
		for i=1, #item.followers do
			GameTooltip:AddLine(item.followers[i], 1, 1, 1)
		end
	end

	GameTooltip:Show()
end

local function updateInProgressMissionsForSelectedCharacter(...)
	if ( isCurrentPlayerSelected() ) then
		module.GarrisonLandingPageReportList_Update(...)
		return;
	end

	local now = time()

	local items = inprogressMissions or {}
	maybeSortInProgressMissionsForSelectedCharacter(items, now)
	local numItems = #items;
	local scrollFrame = GarrisonLandingPageReport.List.listScroll;
	local buttons = scrollFrame.buttons;
	local offset = HybridScrollFrame_GetOffset(scrollFrame);

	if (numItems == 0) then
		GarrisonLandingPageReport.List.EmptyMissionText:SetText(GARRISON_EMPTY_IN_PROGRESS_LIST);
	else
		GarrisonLandingPageReport.List.EmptyMissionText:SetText(nil);
	end

	for i = 1, #buttons do
		local button = buttons[i];
		local index = offset + i; -- adjust index
		local item = items[index];

		if ( item ) then
			button.id = index;
			local bgName;
			if (item.isBuilding or item.canActivate) then
				bgName = "GarrLanding-Building-";
				button.Status:SetText(GARRISON_LANDING_STATUS_BUILDING);
			else
				bgName = "GarrLanding-Mission-";
				button.Status:SetText(nil);
			end
			button.Title:SetText(item.name);

			local isComplete = (now >= item.endDate)
			if ( isComplete ) then
				bgName = bgName.."Complete";
				button.MissionType:SetText(GARRISON_LANDING_BUILDING_COMPLEATE);
				button.MissionType:SetTextColor(YELLOW_FONT_COLOR.r, YELLOW_FONT_COLOR.g, YELLOW_FONT_COLOR.b);
				button.Title:SetWidth(290);
				--[===[@debug@
				button.Seconds:SetText(nil)
				button.Seconds:Hide()
				--@end-debug@]===]
			else
				bgName = bgName.."InProgress";
				button.MissionType:SetTextColor(GARRISON_MISSION_TYPE_FONT_COLOR.r, GARRISON_MISSION_TYPE_FONT_COLOR.g, GARRISON_MISSION_TYPE_FONT_COLOR.b);
				if (item.isBuilding) then
					button.MissionType:SetText(GARRISON_BUILDING_IN_PROGRESS);
				else
					button.MissionType:SetText(item.type);
				end
				button.TimeLeft:SetText(PreLoadLib.tracker.secondsToString(item.endDate - now));
				button.Title:SetWidth(322 - button.TimeLeft:GetWidth());
				--[===[@debug@
				button.Seconds:SetText(item.endDate - now)
				button.Seconds:Show()
				--@end-debug@]===]
			end
			button.MissionTypeIcon:SetAtlas(item.typeAtlas);
			button.MissionTypeIcon:SetShown(not item.isBuilding);
			button.Status:SetShown(not isComplete);
			button.TimeLeft:SetShown(not isComplete);

			button.BG:SetAtlas(bgName, true);
			for i = 1, #button.Rewards do
				button.Rewards[i]:Hide();
			end
			button:Show();
		else
			button:Hide();
		end
	end

	local totalHeight = numItems * scrollFrame.buttonHeight;
	local displayedHeight = #buttons * scrollFrame.buttonHeight;
	HybridScrollFrame_Update(scrollFrame, totalHeight, displayedHeight);
end

local function updateAvailableMissionsForSelectedCharacter(...)
	if ( isCurrentPlayerSelected() ) then
		module.GarrisonLandingPageReportList_UpdateAvailable(...)
		return;
	end

	local items = {};
	local numItems = #items;
	local scrollFrame = GarrisonLandingPageReport.List.listScroll;
	local buttons = scrollFrame.buttons;
	for i = 1, #buttons do
		local button = buttons[i];
		button:Hide();
	end

	GarrisonLandingPageReport.List.EmptyMissionText:SetText(GARRISON_EMPTY_MISSION_LIST);


	local totalHeight = numItems * scrollFrame.buttonHeight;
	local displayedHeight = #buttons * scrollFrame.buttonHeight;
	HybridScrollFrame_Update(scrollFrame, totalHeight, displayedHeight);
end

local function updateMissionsForSelectedCharacter()
	if ( isCurrentPlayerSelected() ) then
		return;
	end

	local self = GarrisonLandingPageReport
	self.InProgress.Text:SetFormattedText(GARRISON_LANDING_IN_PROGRESS, inprogressMissions and #inprogressMissions or 0);
	self.Available.Text:SetFormattedText(GARRISON_LANDING_AVAILABLE, 0);
end

local function showShipmentTooltipForSelectedCharacter(self, ...)
	if ( isCurrentPlayerSelected() ) then
		module.GarrisonLandingPageReportShipment_OnEnter(self, ...)
		return;
	end

	local workorders = PreLoadLib.Globals.Characters[selectedPlayer].workorders
	if (not workorders or #workorders == 0 ) then
		return;
	end

	local now = time()

	local workorder = workorders[self.id]
	local name = workorder.name
	local texture = workorder.texture
	local shipmentCapacity = workorder.shipmentCapacity
	local shipmentsReady = workorder.shipmentsReady
	local shipmentsTotal = workorder.shipmentsTotal
	local creationTime = workorder.creationTime
	local duration = workorder.duration
	local itemName = workorder.itemname
	local buildingID = workorder.buildingID
	local plotID = workorder.plotID

	if ( shipmentsTotal and duration > 0 and now > creationTime + duration ) then -- Adjust ready and creation time based on elapsed time
		shipmentsReady = min(shipmentsTotal, floor(shipmentsReady + (now - creationTime) / duration))
		creationTime = now - (now - creationTime) % duration
	end

	GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
	if (itemName) then
		GameTooltip:SetText(itemName);
	else
		GameTooltip:SetText(name);
	end

	local buildings = PreLoadLib.Globals.Characters[selectedPlayer].buildings or {}
	local building = buildings[plotID]
	if ( building and (building.isBuilding or building.canActivate) ) then
		GameTooltip:AddLine(GARRISON_BUILDING_UNDER_CONSTRUCTION, 1, 1, 1);
	else
		GameTooltip:AddLine(GARRISON_LANDING_SHIPMENT_LABEL, 1, 1, 1);
		GameTooltip:AddLine(" ");

		local shipmentsAvailable = shipmentCapacity;

		if(shipmentsTotal) then
			shipmentsAvailable = shipmentCapacity - shipmentsTotal;
		end

		if shipmentsAvailable > 0 then
			GameTooltip:AddLine(format(GARRISON_LANDING_SHIPMENT_READY_TO_START, shipmentsAvailable) , GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b);
		end

		if (shipmentsReady and shipmentsTotal) then
			if (shipmentsReady == shipmentsTotal) then
				GameTooltip:AddLine(format(GARRISON_LANDING_RETURN, shipmentsTotal), GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b);
			else
				if (creationTime) then
					local timeleftString = PreLoadLib.tracker.secondsToString(creationTime + duration - now)
					GameTooltip:AddLine(format(GARRISON_LANDING_COMPLETED, shipmentsReady, shipmentsTotal) .. " " .. format(GARRISON_LANDING_NEXT, timeleftString), 1, 1, 1)
				else
					GameTooltip:AddLine(format(GARRISON_LANDING_COMPLETED, shipmentsReady, shipmentsTotal), 1, 1, 1)
				end
			end
		end
	end
	GameTooltip:Show();
end

local shipmentUpdateTimer = nil
local function updateShipmentsForSelectedCharacter(self)
	local shipmentIndex = 1;
	local workorders = PreLoadLib.Globals.Characters[selectedPlayer].workorders
	if (not workorders or #workorders == 0 ) then
		self.EmptyString:SetText(L_NO_DATA_AVAILABLE);
		self.EmptyString:Show();
		workorders = {}
	end

	local nextUpdate = 1 * MINUTES
	local now = time()
	for i = 1, #workorders do
		local workorder = workorders[i]
		local name = workorder.name
		local texture = workorder.texture
		local shipmentCapacity = workorder.shipmentCapacity
		local shipmentsReady = workorder.shipmentsReady
		local shipmentsTotal = workorder.shipmentsTotal
		local creationTime = workorder.creationTime
		local duration = workorder.duration
		local buildingID = workorder.buildingID
		local plotID = workorder.plotID

		local shipment = self.Shipments[shipmentIndex];
		if ( not shipment ) then
			return;
		end

		if ( name and shipmentCapacity > 0 ) then
			SetPortraitToTexture(shipment.Icon, texture);
			shipment.id = i
			shipment.Icon:SetDesaturated(true);
			shipment.Name:SetText(name);
			shipment.Done:Hide();
			shipment.Border:Show();
			shipment.BG:Hide();
			shipment.Count:SetText(nil);
			shipment.buildingID = buildingID;
			shipment.plotID = plotID;
			if (shipmentsTotal) then
				if ( duration > 0 and now > creationTime + duration ) then -- Adjust ready and creation time based on elapsed time
					shipmentsReady = min(shipmentsTotal, floor(shipmentsReady + (now - creationTime) / duration))
					creationTime = now - (now - creationTime) % duration
				end

				shipment.Count:SetFormattedText(GARRISON_LANDING_SHIPMENT_COUNT, shipmentsReady, shipmentsTotal);
				if ( shipmentsReady == shipmentsTotal ) then
					shipment.Swipe:SetCooldownUNIX(0, 0);
					shipment.Done:Show();
					shipment.Border:Hide();
				else
					shipment.BG:Show();
					shipment.Swipe:SetCooldownUNIX(creationTime, duration);
					nextUpdate = math.min(nextUpdate, duration - (now - creationTime))
				end
			else
				shipment.Swipe:SetCooldownUNIX(0, 0);
			end
			shipment:Show();
			shipmentIndex = shipmentIndex + 1;
		else
			shipment:Hide();
		end
	end
	for i = shipmentIndex, #self.Shipments do
		self.Shipments[i]:Hide();
	end

	if ( nextUpdate > 0 )  then
		DEBUG("Next shipment update in %d seconds", nextUpdate)
		shipmentUpdateTimer = C_Timer.NewTimer(nextUpdate, function()
			GarrisonLandingPageReport_GetShipments(self)
		end)
	end
end


local function loadShipments(self, ...)
	local cacheAmount = PreLoadLib.tracker.estimateGarrisonCache(selectedPlayer);
	if ( cacheAmount ) then
		self.GarrisonResourceCache.Amount:SetText(cacheAmount);
		self.GarrisonResourceCache:Show();
	else
		self.GarrisonResourceCache:Hide();
	end

	-- Reset shipment cooldown graphic
	for i = 1, #self.Shipments do
		self.Shipments[i].Swipe:SetCooldownUNIX(0, 0);
	end

	-- Cancel update timer
	if ( shipmentUpdateTimer ) then
		shipmentUpdateTimer:Cancel()
		shipmentUpdateTimer = nil
	end

	self.EmptyString:Hide();
	if ( isCurrentPlayerSelected() ) then
		module.GarrisonLandingPageReport_GetShipments(self, ...)
	else
		updateShipmentsForSelectedCharacter(self, ...)
	end
end


local function handleMissionClicksForCharacters(self, button, ...)
	if ( isCurrentPlayerSelected() ) then
		module.GarrisonLandingPageReportMission_OnClick(self, button, ...)
		return;
	end
end

local function initialize()
	DEBUG("   |cffffffffTracker|r initialized");

	-- Create CharacterName fontstring
	GarrisonLandingPageReport.CharacterName = GarrisonLandingPageReport:CreateFontString(nil, "BORDER", "GameFontNormal")
	GarrisonLandingPageReport.CharacterName:SetPoint("TOPLEFT", GarrisonLandingPage.HeaderBar, "BOTTOMLEFT", 26, 13)
	GarrisonLandingPageReport.CharacterName:Show()

	-- Reset seleccted player on LandingPage close
	GarrisonLandingPage:HookScript("OnHide", selectCurrentPlayer)

	-- Hide dropdown on LandingPage clsoe
	GarrisonLandingPageReport:HookScript("OnHide", function()
		if ( UIDropDownMenu_GetCurrentDropDown() == MissionMinderCharacterMenu ) then
			CloseDropDownMenus();
		end
	end);

	-- Track mission progress across multiple characters
	Lib.RegisterCallback("tracker", "MISSION_STARTED", trackMission)
	Lib.RegisterCallback("tracker", "MISSION_COMPLETED", untrackMission)

	-- Identify and cleanup orphaned missions (missions that are tracked, but were already completed outside of the addon)
	--Note: MissionList seems like the logical choice here but using MissionList actually causes us to erroneously
	--remove newly tracked misisons
	GarrisonMissionFrame.MissionTab:HookScript("OnShow", cleanupOrphanedMissions)

	-- Track finished missions (mostly needed for when Mission Completion Orders are used)
	Lib.events:RegisterEvent("tracker", "GARRISON_MISSION_FINISHED", trackFinishedMissions)
	GarrisonMissionFrame:HookScript("OnShow", trackFinishedMissions)

	-- Handle mission button clicks across characters
	module.GarrisonLandingPageReportMission_OnClick = GarrisonLandingPageReportMission_OnClick
	GarrisonLandingPageReportMission_OnClick = handleMissionClicksForCharacters
	local missionButtons = GarrisonLandingPageReport.List.listScroll.buttons
	for i = 1, #missionButtons do
		missionButtons[i]:SetScript("OnClick", GarrisonLandingPageReportMission_OnClick)
	end

	GarrisonLandingPageReport.EmptyString = GarrisonLandingPageReport:CreateFontString(nil, "BORDER", "GameFontHighlightLarge")
	GarrisonLandingPageReport.EmptyString:SetPoint("BOTTOMLEFT", 60, 40)
	GarrisonLandingPageReport.EmptyString:SetPoint("TOPRIGHT", GarrisonLandingPageReport.List, "TOPLEFT", 0, 0)
	GarrisonLandingPageReport.EmptyString:SetTextColor(1, 0.82, 0)
	GarrisonLandingPageReport.EmptyString:Hide()

	-- Create disbaled fonts for mission tabs
	GarrisonLandingPageReport.InProgress:SetDisabledFontObject(GameFontDisableMed2)
	GarrisonLandingPageReport.Available:SetDisabledFontObject(GameFontDisableMed2)

	-- Update Shipments for the selected character
	module.GarrisonLandingPageReport_GetShipments = GarrisonLandingPageReport_GetShipments
	GarrisonLandingPageReport_GetShipments = loadShipments

	-- Update shipment tooltip for the selected character
	module.GarrisonLandingPageReportShipment_OnEnter = GarrisonLandingPageReportShipment_OnEnter
	GarrisonLandingPageReportShipment_OnEnter = showShipmentTooltipForSelectedCharacter
	local shipments = GarrisonLandingPageReport.Shipments
	for i = 1, #shipments do
		shipments[i]:SetScript("OnEnter", GarrisonLandingPageReportShipment_OnEnter);
	end

	-- Update missions for the selected character
	hooksecurefunc("GarrisonLandingPageReportList_UpdateItems", updateMissionsForSelectedCharacter)
	module.GarrisonLandingPageReportList_UpdateAvailable = GarrisonLandingPageReportList_UpdateAvailable
	GarrisonLandingPageReportList_UpdateAvailable = updateAvailableMissionsForSelectedCharacter
	module.GarrisonLandingPageReportList_Update = GarrisonLandingPageReportList_Update
	GarrisonLandingPageReportList_Update = updateInProgressMissionsForSelectedCharacter

	-- Update mission tooltip for the selected character
	module.GarrisonLandingPageReportMission_OnEnter = GarrisonLandingPageReportMission_OnEnter
	GarrisonLandingPageReportMission_OnEnter = updateMissionTooltipForSelectedCharacter
	for i=1, #missionButtons do
		missionButtons[i]:SetScript("OnEnter", GarrisonLandingPageReportMission_OnEnter)
	end

	--[===[@debug@
	for i=1, #missionButtons do
		local secondsRemaining = missionButtons[i]:CreateFontString(nil, "ARTWORK", "GameFontNormal")
		secondsRemaining:SetPoint("BOTTOMRIGHT", missionButtons[i], -4, 4)
		secondsRemaining:SetJustifyH("RIGHT")
		secondsRemaining:SetTextColor(1, 1, 1)
		missionButtons[i].Seconds = secondsRemaining
	end
	--@end-debug@]===]
end

local events = Lib.events
events:RegisterEvent("tracker", "ADDON_LOADED", initialize);
